/*
 * APPR - Automated Profiling and Performance Regression - is an
 * extensible performance profiling framework for OpenOffice.org.
 * 
 * Copyright (C) 2005 Intel Corporation
 * 
 * GNU Lesser General Public License Version 2.1
 * ======================================================================
 * 
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the
 * License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
 * USA
 * 
 * ======================================================================
 * 
 * Hook enabling code for invoke the performance utilities provided by the framework.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "appr_hook.h"

int appr_hook_recv_command(int comm, struct apprcmd *command, struct apprbuf *buffer)
{
	int i, value;
	int length;

	/* check the file descriptor is valid */
	if (comm < 0) {
		return -1;
	}

	/* check the structure pointer is valid */
	if ((command == NULL) || (buffer == NULL)) {
		return -1;
	}

	length = buffer->end - buffer->start;
	if (length < 0) {
		buffer->end = buffer->start = 0;
		return -1;
	}

	/* get the header of the command */
	command->expect_length = 3;
	command->command_length = 0;
	while (length < command->expect_length) {
		if (buffer->start != 0) {
			buffer->start = 0;
			buffer->end = length;
			if (length > 0) {
				memcpy(buffer->res, &(buffer->res[buffer->start]), length);
			}
		}

		value = read(comm, &(buffer->res[buffer->end]), (sizeof(buffer->res) - buffer->end));
		if (value <= 0) {
			return -1;
		}

		length += value;
		buffer->end += length;
	}

	/* get the rest of the command */
	if (buffer->res[1] < command->expect_length) {
		return -1;
	}

	command->expect_length = buffer->res[1] - command->command_length;
	while (length < command->expect_length) {
		if (buffer->start + command->expect_length > sizeof(buffer->res)) {
			memcpy(&(command->res[command->command_length]), &(buffer->res[buffer->start]), length);
			command->command_length += length;
			length = buffer->end = buffer->start = 0;
		}

		value = read(comm, &(buffer->res[buffer->end]), (sizeof(buffer->res) - buffer->end));
		if (value <= 0) {
			return -1;
		}

		length += value;
		buffer->end += length;
	}

	memcpy(&(command->res[command->command_length]), &(buffer->res[buffer->start]), length);
	command->command_length += length;
	command->expect_length = 0;
	buffer->start += length;

	/* decode the command */
	command->code = command->res[0];
	command->size = command->res[1];
	command->desc = NULL;
	switch(command->code) {
	case APPR_HOOK_REQUEST:
	case APPR_HOOK_REPLY:
		command->argc = 1;
		command->argv = &(command->res[2]);
		break;
	case APPR_HOOK_CLOSE:
		command->argc = 0;
		command->argv = NULL;
		break;
	default:
		command->failure = 1;
		return -1;
	}

	i = command->argc + 2;
	if (command->res[i] == APPR_HOOK_DESCRIPTION) {
		command->desc = &(command->res[++i]);
	}

	if (command->res[command->size - 1] == APPR_HOOK_TOKEN) {
		command->failure = 0;
	} else {
		command->failure = 1;
	}
	command->res[command->size - 1] = '\0';

#ifdef __validation
	fprintf(stderr, "Receive hook command (size 0x%02x): 0x%02x [", command->size, command->code);
	for (i = 0; i < command->argc; i++) {
		fprintf(stderr, " 0x%02x", command->argv[i]);
	}
	fprintf(stderr, " ] (%s)\n", command->desc);
#endif // __validation

	return 0;
}

int appr_hook_send_command(int comm, struct apprcmd *command, unsigned char code, char *desc, ...)
{
	int i, value;
	int length, remain_length;
	va_list arg_ptr;

	/* check the file descriptor is valid */
	if (comm < 0) {
		return -1;
	}

	/* check the structure pointer is valid */
	if (command == NULL) {
		return -1;
	}

	command->code = command->res[0] = code;
	switch(code) {
	case APPR_HOOK_REQUEST:
	case APPR_HOOK_REPLY:
		command->argc = 1;
		command->argv = &(command->res[2]);
		break;
	case APPR_HOOK_CLOSE:
		command->argc = 0;
		command->argv = NULL;
		break;
	default:
		command->failure = 1;
		return -1;
	}

	va_start(arg_ptr, desc);
	for (i = 0; i < command->argc; i++) {
		command->argv[i] = (unsigned char) va_arg(arg_ptr, unsigned int);
	}
	va_end(arg_ptr);

	command->size = command->argc + 3;
	i = command->argc + 2;

	if (desc != NULL) {
		command->res[i++] = APPR_HOOK_DESCRIPTION;
		command->desc = &(command->res[i]);

		length = strlen(desc);
		remain_length = sizeof(command->res) - i - 1;
		if (length > remain_length) {
			length = remain_length;
		}

		memcpy(command->desc, desc, length);
		command->size += length + 1;
	} else {
		command->desc = NULL;
	}

	command->res[command->size - 1] = APPR_HOOK_TOKEN;
	command->res[1] = command->size;
	command->failure = 0;

#ifdef __validation
	fprintf(stderr, "Send hook command (size 0x%02x): 0x%02x [", command->res[1], command->res[0]);
	for (i = 0; i < command->argc; i++) {
		fprintf(stderr, " 0x%02x", command->argv[i]);
	}
	fprintf(stderr, " ] (%s)\n", command->desc);
#endif // __validation

	value = write(comm, command->res, command->size);

	if (value < command->size) {
		return -1;
	} else {
		return 0;
	}
}
