/*
 * APPR - Automated Profiling and Performance Regression - is an
 * extensible performance profiling framework for OpenOffice.org.
 * 
 * Copyright (C) 2005 Intel Corporation
 * 
 * GNU Lesser General Public License Version 2.1
 * ======================================================================
 * 
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the
 * License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
 * USA
 * 
 * ======================================================================
 * 
 * Hook agent utility for APPR (Automated Performance Profiling Regression).
 * To be invoked from target program by function or shell script.
 * It will interact with hook daemon to handle the hook requests.
 *
 */

// #define __validation

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <dlfcn.h>

#include <regex.h>

#include "appr_hook.h"

static int enable = FALSE;
static int initialized = FALSE;
static int enable_vtune = FALSE;
static int vtune_paused;
static int enable_strace = FALSE;
static int enable_callgrind = FALSE;

// vtune_paused represent the VTune running state prior to get the instrument point
static regex_t appr_hook_regex;
static regex_t appr_hook_pause_regex;
static regex_t appr_hook_resume_regex;

static void (*vtune_pause)(void);
static void (*vtune_resume)(void);

static int comm = INVALID;

static int appr_hook_env()
{
	char *buf;
	int value;

	void *dll_ptr;

	buf = getenv(APPR_HOOK_ENVVAR_PROG);
	if (buf == NULL) {
		enable = FALSE;
		return 0;
	}

	buf = getenv(APPR_HOOK_ENVVAR_OOO_RTL);
	if (buf == NULL) {
		enable = FALSE;
		return 0;
	}

	enable = TRUE;

	/* initialize hook filter regular expression */
	buf = getenv(APPR_HOOK_ENVVAR_REGEX);
	if (buf != NULL) {
		regcomp(&appr_hook_regex, buf, REG_EXTENDED|REG_NOSUB);
#ifdef __validation
		fprintf(stderr, "hook filter regex: %s\n", buf);
#endif // __validation
	} else {
		regcomp(&appr_hook_regex, REGEX_MATCH, REG_EXTENDED|REG_NOSUB);
#ifdef __validation
		fprintf(stderr, "hook filter regex: %s (NULL)\n", REGEX_MATCH);
#endif // __validation
	}

	/* initialize pause regular expression */
	buf = getenv(APPR_HOOK_ENVVAR_PAUSE_REGEX);
	if (buf != NULL) {
		regcomp(&appr_hook_pause_regex, buf, REG_EXTENDED|REG_NOSUB);
#ifdef __validation
		fprintf(stderr, "pause regex: %s\n", buf);
#endif // __validation
	} else {
		regcomp(&appr_hook_pause_regex, REGEX_NONMATCH, REG_EXTENDED|REG_NOSUB);
#ifdef __validation
		fprintf(stderr, "pause regex: %s\n", REGEX_NONMATCH);
#endif // __validation
	}

	/* initialize resume regular expression */
	buf = getenv(APPR_HOOK_ENVVAR_RESUME_REGEX);
	if (buf != NULL) {
		regcomp(&appr_hook_resume_regex, buf, REG_EXTENDED|REG_NOSUB);
#ifdef __validation
		fprintf(stderr, "resume regex: %s\n", buf);
#endif // __validation
	} else {
		regcomp(&appr_hook_resume_regex, REGEX_NONMATCH, REG_EXTENDED|REG_NOSUB);
#ifdef __validation
		fprintf(stderr, "resume regex: %s\n", REGEX_NONMATCH);
#endif // __validation
	}

	/* check whether to enable VTune support, if so initialize VTune API
	through loading dynamic link library
	*/
	buf = getenv(APPR_HOOK_ENVVAR_VTUNE);
	if (buf != NULL) {
		dll_ptr = dlopen(LIB_VTUNEAPI, RTLD_LAZY);
		if (dll_ptr != NULL) {
			vtune_pause = dlsym(dll_ptr, FNT_VTUNEPAUSE);
			vtune_resume = dlsym(dll_ptr, FNT_VTUNERESUME);

			if ((vtune_pause != NULL) && (vtune_resume != NULL)) {
				enable_vtune = TRUE;
				fprintf(stderr, "VTune Api is available.\n");
			} else {
				enable_vtune = FALSE;
				fprintf(stderr, "VTune control feature disabled: [%s].\n", (int)dlerror());
			}
		} else {
			enable_vtune = FALSE;
			fprintf(stderr, "VTune control feature disabled: [%s].\n", (int)dlerror());
		}
	} else {
		enable_vtune = FALSE;
	}

	if (enable_vtune == TRUE) {

		buf = getenv(APPR_HOOK_ENVVAR_VTUNE_STARTPAUSED);
		if (buf != NULL) {
			vtune_paused = TRUE;
			fprintf(stderr, "VTune is paused.\n");
		} else {
			vtune_paused = FALSE;
			vtune_pause();
			fprintf(stderr, "VTune is not paused.\n");
		}
	}

	buf = getenv(APPR_HOOK_ENVVAR_STRACE);
	if (buf != NULL) {
		enable_strace = TRUE;
	} else {
		enable_strace = FALSE;
	}

	buf = getenv(APPR_HOOK_ENVVAR_CALLGRIND);
	if (buf != NULL) {
		enable_callgrind = TRUE;
	} else {
		enable_callgrind = FALSE;
	}

	/* initialize the communication channel to hook daemon */
	if (comm == INVALID) {
		struct sockaddr_un sockaddr;

		comm = socket(AF_UNIX, SOCK_STREAM, PF_UNIX);
		if (comm == INVALID) {
			enable = FALSE;
			return -1;
		}

		sockaddr.sun_family = AF_UNIX;

		buf = getenv(APPR_HOOK_ENVVAR_DAEMON_ID);
		if (buf != NULL) {
			memcpy(sockaddr.sun_path, buf, sizeof(sockaddr.sun_path));
		} else {
			memcpy(sockaddr.sun_path, APPR_HOOK_DAEMON_ID, sizeof(sockaddr.sun_path));
		}

		value = connect(comm, (struct sockaddr *)&sockaddr, sizeof(sockaddr));
		if (value < 0) {
			close(comm);
			comm = INVALID;
			enable = FALSE;
			return -1;
		}
	}

	return 0;
}

static int appr_hook_agent_proc(unsigned int command, unsigned int type, const char *msg, va_list arg_ptr)
{
	int value, pause_match, resume_match;
	unsigned char message[APPR_HOOK_BUFFER_SIZE];

	struct apprcmd appr_command;
	struct apprbuf appr_buffer;

	/* initialize state and variables for the first time */
	if (initialized == FALSE) {
		initialized = TRUE;
		appr_hook_env();
#ifdef __validation
		fprintf(stderr, "Enable: ", initialized);
		if (enable == TRUE) {
			fprintf(stderr, "Yes (");
	
			if (enable_vtune == TRUE) {
				fprintf(stderr, " VTune");
			}
	
			if (enable_strace == TRUE) {
				fprintf(stderr, " strace");
			}
	
			if (enable_callgrind == TRUE) {
				fprintf(stderr, " Callgrind");
			}
		
			fprintf(stderr, " ).\n");
		} else {
			fprintf(stderr, "No.\n");
		}
#endif // __validation
	}

	if (enable == TRUE) {

		/* pause VTune sampling at the point as early as possible */
		if (enable_vtune == TRUE) {
			if (vtune_paused == FALSE) {
				vtune_pause();
#ifdef __validation
				fprintf(stderr, "VTune is paused:[%s]\n", msg);
#endif // __validation
			}
		}

#ifdef __validation
		fprintf(stderr, "Type %u: (", type);
		fprintf(stderr, msg, arg_ptr);
		fprintf(stderr, ")\n");
#endif // __validation

		/* check whether current instrument is filtered */
		value = regexec(&appr_hook_regex, (const char *)msg, 0, NULL, 0);
		if (value != 0) {
			return 0;
		}

		pause_match = regexec(&appr_hook_pause_regex, (const char *)msg, 0, NULL, 0);
		resume_match = regexec(&appr_hook_resume_regex, (const char *)msg, 0, NULL, 0);

		switch (command) {
		case APPR_HOOK_CLOSE:
			/* only send out the command */
			value = appr_hook_send_command(comm, &appr_command, APPR_HOOK_CLOSE, NULL);
			break;
	
		case APPR_HOOK_REQUEST:
/*
			if (type >= APPR_HOOK_TYPE_NUMBERS) {
				type = APPR_HOOK_PHASESTARTSTOP;
			}
*/
			vsnprintf((char *)message, sizeof(message), msg, arg_ptr);
		
			/* construct and send out hook request */
			value = appr_hook_send_command(comm, &appr_command, APPR_HOOK_REQUEST, (char *)message, type);
			if (value < 0) {
				break;
			}
		
			/* wait for response from hook daemon */
			appr_buffer.end = appr_buffer.start = 0;
			value = 1;
			while (value > 0) {
				value = appr_hook_recv_command(comm, &appr_command, &appr_buffer);
				if (value < 0) {
					break;
				}
		
				switch(appr_command.code) {
				case APPR_HOOK_REPLY:
					if (appr_command.argv[0] == 0) {
						value = 0;
					} else {
						value = -1;
					}
					break;
				default:
					value = 1; /* wait for APPR_HOOK_REPLY */
				}
			}
		}

		/* resume VTune sampling at the point as late as possible */
		if (enable_vtune == TRUE) {
#ifdef __validation
			fprintf(stderr, "vtune is paused: ");
#endif // __validation
	
			// assume the vtune activity should be paused at the beginning of this instrument
			if (vtune_paused == FALSE) {
	
#ifdef __validation
				fprintf(stderr, " false\n");
#endif // __validation
	
				// pause regular expression matched -> donot resume vtune activity
				if (pause_match == 0) {
					vtune_paused = TRUE;
					fprintf(stderr, "VTune is paused:[%s]\n", msg);
				} else {
					vtune_resume();
#ifdef __validation
					fprintf(stderr, "VTune is resumed:[%s]\n", msg);
#endif // __validation
				}
	
			} else {
	
#ifdef __validation
				fprintf(stderr, " true\n");
#endif // __validation
	
				// resume regular expression matched -> resume vtune activity
				if (resume_match == 0) {
					vtune_resume();
					vtune_paused = FALSE;
					fprintf(stderr, "VTune is resumed:[%s]\n", msg);
				}
	
			}
		}
	}

	return value;
}

int appr_hook_close()
{
	return appr_hook_agent_proc(APPR_HOOK_CLOSE, APPR_HOOK_PHASENCARE, NULL, NULL);
}

int appr_hook_agent(unsigned int type, const char *msg, ...)
{
	int value;
	va_list arg_ptr;

	va_start(arg_ptr, msg);
	value = appr_hook_agent_proc(APPR_HOOK_REQUEST, type, msg, arg_ptr);
	va_end(arg_ptr);

	return value;
}

#ifdef __standalone
static int usage(int argc, char *argv[])
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "\t%s <command> [options ...]\n", argv[0]);
	fprintf(stderr, "\tcommand: request, close\n");
	return 0;
}

int main(int argc, char *argv[])
{
	int value;
	unsigned int type;
	char **arg_ptr;

	if (argc < 2) {
		usage(argc, argv);
		return -1;
	}

	if (strcmp(argv[1], "request") == 0) {
		if (argc < 4) {
			fprintf(stderr, "Usage:\n");
			fprintf(stderr, "\t%s request <type> <message> ...\n", argv[0]);
			return -1;
		} else {
			arg_ptr = &(argv[1]);
		}
	} else if (strcmp(argv[1], "close") == 0) {

		return appr_hook_close();

#ifdef __validation
	} else if (strcmp(argv[1], "request-ns") == 0) {
		if (argc < 4) {
			fprintf(stderr, "Usage (used in validation for non-standalone mode):\n");
			fprintf(stderr, "\t%s request-ns <type> <message> ...\n", argv[0]);
			return -1;
		} else {
			arg_ptr = &(argv[1]);
		}
#endif // __validation

	} else {
		usage(argc, argv);
		return -1;
	}

	/* validation for non-standalone usage */

	type = atoi(arg_ptr[1]);

	if (argc == 3) {
		value = appr_hook_agent_proc(APPR_HOOK_REQUEST, type, arg_ptr[2], NULL);
	} else {
		value = appr_hook_agent_proc(APPR_HOOK_REQUEST, type, arg_ptr[2], (va_list)&arg_ptr[3]);
	}

	return value;
}
#endif // __standalone

