/*
 * APPR - Automated Profiling and Performance Regression - is an
 * extensible performance profiling framework for OpenOffice.org.
 * 
 * Copyright (C) 2005 Intel Corporation
 * 
 * GNU Lesser General Public License Version 2.1
 * ======================================================================
 * 
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the
 * License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
 * USA
 * 
 * ======================================================================
 * 
 * Hook daemon utility for APPR (Automated Performance Profiling Regression).
 * To receive hook requests from hook agent and then invoke APPR hook program.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <regex.h>

#include "appr_hook.h"

static int initialized = FALSE;
static char *appr_hook_program = NULL;
static regex_t appr_hook_regex;

const char *appr_hook_type[] = {
	" start ",
	" stop ",
	" startstop ",
	" ncare ",
	NULL
};

// Globalize these to help with cleanup
int sock_sche = -1;
int sock_serv = -1;
char sockname[256];

void cleanup(void) {
	struct stat filebuf;

	/* close socket */
	if (sock_sche != -1) {
		close(sock_sche);
		sock_sche = -1;
	}

	if (sock_serv != -1) {
		close(sock_serv);
		sock_serv = -1;
	}

	/* remove socket file */
	if(stat(sockname, &filebuf) == 0)
	{
		unlink(sockname);
	}
}

int main(int argc, char *argv[])
{
	int value;
	char *buf;

	int addr_len;
	struct sockaddr_un sockaddr_daemon, sockaddr_agent;

	struct apprcmd command;
	struct apprbuf buffer;

	char syscommand[1000];

	/* Build the socket name */
	buf = getenv(APPR_HOOK_ENVVAR_DAEMON_ID);
	if (buf != NULL) {
		strncpy(sockname, buf, sizeof(sockname));
	} else {
		strncpy(sockname, APPR_HOOK_DAEMON_ID, sizeof(sockname));
	}

	/* Register function to be called at exit */
	atexit(cleanup);

	/* initialize */
	if (initialized == FALSE) {
		initialized = TRUE;
		appr_hook_program = getenv(APPR_HOOK_ENVVAR_PROG);
		if (appr_hook_program == NULL) {
			return -1;
		}

		if (strlen(appr_hook_program) >= sizeof(syscommand)) {
			return -1;
		}
	}

	/* start daemon to listen to request from hook agent */
	sock_sche = socket(AF_UNIX, SOCK_STREAM, PF_UNIX);
	if (sock_sche == INVALID) {
		return -1;
	}

	sockaddr_daemon.sun_family = AF_UNIX;
	memcpy(sockaddr_daemon.sun_path, sockname, sizeof(sockaddr_daemon.sun_path));

	value = bind(sock_sche, (struct sockaddr *)&sockaddr_daemon, sizeof(sockaddr_daemon));
	if (value < 0) {
		return -1;
	}

	value = listen(sock_sche, 1);
	if (value < 0) {
		return -1;
	}

	while (1) {

		/* accept a connection from agent */
		sock_serv = accept(sock_sche, (struct sockaddr *)&sockaddr_agent, (socklen_t *)&addr_len);
		if (sock_serv == INVALID) {
			usleep(100);
			continue;
		}

		/* handle the hook command from agent */
		while (1) {
			value = appr_hook_recv_command(sock_serv, &command, &buffer);
			if (value < 0) {
				close(sock_serv);
				break;
			}

			switch(command.code) {
			case APPR_HOOK_REQUEST:
				/* invoke appr_hook utility */
				strncpy(syscommand, appr_hook_program, sizeof(syscommand));

				strncat(syscommand, appr_hook_type[command.argv[0]], sizeof(syscommand) - strlen(syscommand));

				if (command.desc != NULL) {
					strncat(syscommand, "\"", sizeof(syscommand) - strlen(syscommand));
					strncat(syscommand, (const char *)command.desc, sizeof(syscommand) - strlen(syscommand));
					strncat(syscommand, "\"", sizeof(syscommand) - strlen(syscommand));
				}

#ifdef __validation
				fprintf(stderr, "Exec hook command (%s)\n", syscommand);
#endif // __validation

				value = system(syscommand);

				/* nofity agent the request result */
				if (value == 0) {
					appr_hook_send_command(sock_serv, &command, APPR_HOOK_REPLY, "succeed", 0);
				} else {
					appr_hook_send_command(sock_serv, &command, APPR_HOOK_REPLY, "fail", -1);
				}

				break;

			case APPR_HOOK_CLOSE:
				return 0;

			default:
				break;
			}

		}
	}
	
	return 0;
}

